/*
 * OPC-UA code representation of MOD-IOs connected to a Lime2
 */

#include <open62541/server.h>

void addIntegerVariableNode(UA_Server *server, char *node_id, char *node_description)
{
    UA_Int32 myInteger = 0;
    UA_NodeId parentNodeId = UA_NODEID_NUMERIC(0, UA_NS0ID_OBJECTSFOLDER);
    UA_NodeId parentReferenceNodeId = UA_NODEID_NUMERIC(0, UA_NS0ID_ORGANIZES);

    UA_VariableAttributes attr0 = UA_VariableAttributes_default;
    UA_Variant_setScalar(&attr0.value, &myInteger, &UA_TYPES[UA_TYPES_INT32]);
    attr0.description = UA_LOCALIZEDTEXT("en-US", node_description);
    attr0.displayName = UA_LOCALIZEDTEXT("en-US", node_description);
    attr0.dataType = UA_TYPES[UA_TYPES_INT32].typeId;
    attr0.accessLevel = UA_ACCESSLEVELMASK_READ | UA_ACCESSLEVELMASK_WRITE;
    UA_NodeId myIntegerNodeId0 = UA_NODEID_STRING(1, node_id);
    UA_QualifiedName myIntegerName0 = UA_QUALIFIEDNAME(1, node_description);
    UA_Server_addVariableNode(server, myIntegerNodeId0, parentNodeId,
                              parentReferenceNodeId, myIntegerName0,
                              UA_NODEID_NUMERIC(0, UA_NS0ID_BASEDATAVARIABLETYPE), attr0, NULL, NULL);
}

void addUIntegerVariableReadNode(UA_Server *server, char *node_id, char *node_description)
{
    UA_UInt32 myInteger = 0;
    UA_NodeId parentNodeId = UA_NODEID_NUMERIC(0, UA_NS0ID_OBJECTSFOLDER);
    UA_NodeId parentReferenceNodeId = UA_NODEID_NUMERIC(0, UA_NS0ID_ORGANIZES);

    UA_VariableAttributes attr0 = UA_VariableAttributes_default;
    UA_Variant_setScalar(&attr0.value, &myInteger, &UA_TYPES[UA_TYPES_UINT32]);
    attr0.description = UA_LOCALIZEDTEXT("en-US", node_description);
    attr0.displayName = UA_LOCALIZEDTEXT("en-US", node_description);
    attr0.dataType = UA_TYPES[UA_TYPES_UINT32].typeId;
    attr0.accessLevel = UA_ACCESSLEVELMASK_READ;
    UA_NodeId myIntegerNodeId0 = UA_NODEID_STRING(1, node_id);
    UA_QualifiedName myIntegerName0 = UA_QUALIFIEDNAME(1, node_description);
    UA_Server_addVariableNode(server, myIntegerNodeId0, parentNodeId,
                              parentReferenceNodeId, myIntegerName0,
                              UA_NODEID_NUMERIC(0, UA_NS0ID_BASEDATAVARIABLETYPE), attr0, NULL, NULL);
}

void addBooleanVariableReadNode(UA_Server *server, char *node_id, char *node_description)
{
    UA_Boolean myBoolean = false;
    UA_NodeId parentNodeId = UA_NODEID_NUMERIC(0, UA_NS0ID_OBJECTSFOLDER);
    UA_NodeId parentReferenceNodeId = UA_NODEID_NUMERIC(0, UA_NS0ID_ORGANIZES);

    UA_VariableAttributes attr0 = UA_VariableAttributes_default;
    UA_Variant_setScalar(&attr0.value, &myBoolean, &UA_TYPES[UA_TYPES_BOOLEAN]);
    attr0.description = UA_LOCALIZEDTEXT("en-US", node_description);
    attr0.displayName = UA_LOCALIZEDTEXT("en-US", node_description);
    attr0.dataType = UA_TYPES[UA_TYPES_BOOLEAN].typeId;
    attr0.accessLevel = UA_ACCESSLEVELMASK_READ;
    UA_NodeId myIntegerNodeId0 = UA_NODEID_STRING(1, node_id);
    UA_QualifiedName myIntegerName0 = UA_QUALIFIEDNAME(1, node_description);
    UA_Server_addVariableNode(server, myIntegerNodeId0, parentNodeId,
                              parentReferenceNodeId, myIntegerName0,
                              UA_NODEID_NUMERIC(0, UA_NS0ID_BASEDATAVARIABLETYPE), attr0, NULL, NULL);
}

static void addVariable(UA_Server *server)
{
    /* 
     * Create all variables representing MOD-IO's relays
     */
    int i,n;;
    int length = getI2CSlaveListLength();
    char *node_id = malloc(sizeof(char) * 10);
    char *node_title = malloc(sizeof(char) * 30);

    for (i = 0; i < length; i++)
    {
         for (n = 0; n < DEFAULT_RELAY_COUNT; n++)
         {
             // add relay nodes
             sprintf(node_id, "i2c%d.relay%d", i, n);
             sprintf(node_title, "I2C%d / Relay %d", i, n);
             addIntegerVariableNode(server, node_id, node_title);
             // add digital in
             sprintf(node_id, "i2c%d.in%d", i, n);
             sprintf(node_title, "I2C%d / Digital Input %d", i, n);
             addBooleanVariableReadNode(server, node_id, node_title);
             // add analog in
             sprintf(node_id, "i2c%d.ain%d", i, n);
             sprintf(node_title, "I2C%d / Analog Input %d", i, n);
             addUIntegerVariableReadNode(server, node_id, node_title);
         }
    }
}

/* Connect to variables to physical relays, analog and digital inputs.
 *
 */
static void beforeReadTime(UA_Server *server,
                           const UA_NodeId *sessionId, void *sessionContext,
                           const UA_NodeId *nodeid, void *nodeContext,
                           const UA_NumericRange *range, const UA_DataValue *data)
{
    // nothing to do as not needed.
}

// Analong Inputs
static void beforeReadTimeAnalongIn(int addr,uint8_t read_addr, const UA_DataValue *data) {
    /*
     * Generic handler for reading from physicall analong input.
     */
    int *data_input = 0;
    if (!I2C_VIRTUAL_MODE) {
      getAnalogInputStateAIN(addr, &data_input, read_addr);
      if (data->value.type == &UA_TYPES[UA_TYPES_UINT32])
      {
        *(UA_UInt32 *)data->value.data = *data_input;
      }
    }
}

// I2C0
static void beforeReadTimeI2C0Ain0(UA_Server *server,
                                   const UA_NodeId *sessionId, void *sessionContext,
                                   const UA_NodeId *nodeid, void *nodeContext,
                                   const UA_NumericRange *range, const UA_DataValue *data){
    beforeReadTimeAnalongIn(I2C_SLAVE_ADDR_LIST[0], 0x30, data);}

static void beforeReadTimeI2C0Ain1(UA_Server *server,
                                   const UA_NodeId *sessionId, void *sessionContext,
                                   const UA_NodeId *nodeid, void *nodeContext,
                                   const UA_NumericRange *range, const UA_DataValue *data){
    beforeReadTimeAnalongIn(I2C_SLAVE_ADDR_LIST[0], 0x31, data);}

static void beforeReadTimeI2C0Ain2(UA_Server *server,
                                   const UA_NodeId *sessionId, void *sessionContext,
                                   const UA_NodeId *nodeid, void *nodeContext,
                                   const UA_NumericRange *range, const UA_DataValue *data){
    beforeReadTimeAnalongIn(I2C_SLAVE_ADDR_LIST[0], 0x32, data);}

static void beforeReadTimeI2C0Ain3(UA_Server *server,
                                   const UA_NodeId *sessionId, void *sessionContext,
                                   const UA_NodeId *nodeid, void *nodeContext,
                                   const UA_NumericRange *range, const UA_DataValue *data){
    beforeReadTimeAnalongIn(I2C_SLAVE_ADDR_LIST[0], 0x33, data);}

// I2C1
static void beforeReadTimeI2C1Ain0(UA_Server *server,
                                   const UA_NodeId *sessionId, void *sessionContext,
                                   const UA_NodeId *nodeid, void *nodeContext,
                                   const UA_NumericRange *range, const UA_DataValue *data){
    beforeReadTimeAnalongIn(I2C_SLAVE_ADDR_LIST[1], 0x30, data);}

static void beforeReadTimeI2C1Ain1(UA_Server *server,
                                   const UA_NodeId *sessionId, void *sessionContext,
                                   const UA_NodeId *nodeid, void *nodeContext,
                                   const UA_NumericRange *range, const UA_DataValue *data){
   beforeReadTimeAnalongIn(I2C_SLAVE_ADDR_LIST[1], 0x31, data);}

static void beforeReadTimeI2C1Ain2(UA_Server *server,
                                   const UA_NodeId *sessionId, void *sessionContext,
                                   const UA_NodeId *nodeid, void *nodeContext,
                                   const UA_NumericRange *range, const UA_DataValue *data){
    beforeReadTimeAnalongIn(I2C_SLAVE_ADDR_LIST[1], 0x32, data);}

static void beforeReadTimeI2C1Ain3(UA_Server *server,
                                   const UA_NodeId *sessionId, void *sessionContext,
                                   const UA_NodeId *nodeid, void *nodeContext,
                                   const UA_NumericRange *range, const UA_DataValue *data){
    beforeReadTimeAnalongIn(I2C_SLAVE_ADDR_LIST[1], 0x33, data);}

// I2C2
static void beforeReadTimeI2C2Ain0(UA_Server *server,
                                   const UA_NodeId *sessionId, void *sessionContext,
                                   const UA_NodeId *nodeid, void *nodeContext,
                                   const UA_NumericRange *range, const UA_DataValue *data){
    beforeReadTimeAnalongIn(I2C_SLAVE_ADDR_LIST[2], 0x30, data);}

static void beforeReadTimeI2C2Ain1(UA_Server *server,
                                   const UA_NodeId *sessionId, void *sessionContext,
                                   const UA_NodeId *nodeid, void *nodeContext,
                                   const UA_NumericRange *range, const UA_DataValue *data){
   beforeReadTimeAnalongIn(I2C_SLAVE_ADDR_LIST[2], 0x31, data);}

static void beforeReadTimeI2C2Ain2(UA_Server *server,
                                   const UA_NodeId *sessionId, void *sessionContext,
                                   const UA_NodeId *nodeid, void *nodeContext,
                                   const UA_NumericRange *range, const UA_DataValue *data){
    beforeReadTimeAnalongIn(I2C_SLAVE_ADDR_LIST[2], 0x32, data);}

static void beforeReadTimeI2C2Ain3(UA_Server *server,
                                   const UA_NodeId *sessionId, void *sessionContext,
                                   const UA_NodeId *nodeid, void *nodeContext,
                                   const UA_NumericRange *range, const UA_DataValue *data){
    beforeReadTimeAnalongIn(I2C_SLAVE_ADDR_LIST[2], 0x33, data);}

// I2C3
static void beforeReadTimeI2C3Ain0(UA_Server *server,
                                   const UA_NodeId *sessionId, void *sessionContext,
                                   const UA_NodeId *nodeid, void *nodeContext,
                                   const UA_NumericRange *range, const UA_DataValue *data){
    beforeReadTimeAnalongIn(I2C_SLAVE_ADDR_LIST[3], 0x30, data);}

static void beforeReadTimeI2C3Ain1(UA_Server *server,
                                   const UA_NodeId *sessionId, void *sessionContext,
                                   const UA_NodeId *nodeid, void *nodeContext,
                                   const UA_NumericRange *range, const UA_DataValue *data){
   beforeReadTimeAnalongIn(I2C_SLAVE_ADDR_LIST[3], 0x31, data);}

static void beforeReadTimeI2C3Ain2(UA_Server *server,
                                   const UA_NodeId *sessionId, void *sessionContext,
                                   const UA_NodeId *nodeid, void *nodeContext,
                                   const UA_NumericRange *range, const UA_DataValue *data){
    beforeReadTimeAnalongIn(I2C_SLAVE_ADDR_LIST[3], 0x32, data);}

static void beforeReadTimeI2C3Ain3(UA_Server *server,
                                   const UA_NodeId *sessionId, void *sessionContext,
                                   const UA_NodeId *nodeid, void *nodeContext,
                                   const UA_NumericRange *range, const UA_DataValue *data){
    beforeReadTimeAnalongIn(I2C_SLAVE_ADDR_LIST[3], 0x33, data);}


// Digital Inputs
static void beforeReadTimeDigitalIn(int addr,
                                    int input_number,
                                    const UA_DataValue *data){
    /*
     * Generic handler for reading and updating a Digital In.
     */
    char *data_input = 0;
    if (!I2C_VIRTUAL_MODE) {
      getDigitalInputState(addr, &data_input);
      if ((*data_input) & (1UL << input_number))
      {
        if (data->value.type == &UA_TYPES[UA_TYPES_BOOLEAN])
        {
            *(UA_Boolean *)data->value.data = true;
        }
      }
      else
      {
        if (data->value.type == &UA_TYPES[UA_TYPES_BOOLEAN])
        {
            *(UA_Boolean *)data->value.data = false;
        }
      }
    }
}

// I2C0
static void beforeReadTimeI2C0In0(UA_Server *server,
                                  const UA_NodeId *sessionId, void *sessionContext,
                                  const UA_NodeId *nodeid, void *nodeContext,
                                  const UA_NumericRange *range, const UA_DataValue *data){
    beforeReadTimeDigitalIn(I2C_SLAVE_ADDR_LIST[0], 0, data);}

static void beforeReadTimeI2C0In1(UA_Server *server,
                                  const UA_NodeId *sessionId, void *sessionContext,
                                  const UA_NodeId *nodeid, void *nodeContext,
                                  const UA_NumericRange *range, const UA_DataValue *data){
    beforeReadTimeDigitalIn(I2C_SLAVE_ADDR_LIST[0], 1, data);}

static void beforeReadTimeI2C0In2(UA_Server *server,
                                  const UA_NodeId *sessionId, void *sessionContext,
                                  const UA_NodeId *nodeid, void *nodeContext,
                                  const UA_NumericRange *range, const UA_DataValue *data){
    beforeReadTimeDigitalIn(I2C_SLAVE_ADDR_LIST[0], 2, data);}

static void beforeReadTimeI2C0In3(UA_Server *server,
                                  const UA_NodeId *sessionId, void *sessionContext,
                                  const UA_NodeId *nodeid, void *nodeContext,
                                  const UA_NumericRange *range, const UA_DataValue *data){
    beforeReadTimeDigitalIn(I2C_SLAVE_ADDR_LIST[0], 3, data);}

// I2C1
static void beforeReadTimeI2C1In0(UA_Server *server,
                                  const UA_NodeId *sessionId, void *sessionContext,
                                  const UA_NodeId *nodeid, void *nodeContext,
                                  const UA_NumericRange *range, const UA_DataValue *data){
    beforeReadTimeDigitalIn(I2C_SLAVE_ADDR_LIST[1], 0, data);}

static void beforeReadTimeI2C1In1(UA_Server *server,
                                  const UA_NodeId *sessionId, void *sessionContext,
                                  const UA_NodeId *nodeid, void *nodeContext,
                                  const UA_NumericRange *range, const UA_DataValue *data){
    beforeReadTimeDigitalIn(I2C_SLAVE_ADDR_LIST[1], 1, data);}

static void beforeReadTimeI2C1In2(UA_Server *server,
                                  const UA_NodeId *sessionId, void *sessionContext,
                                  const UA_NodeId *nodeid, void *nodeContext,
                                  const UA_NumericRange *range, const UA_DataValue *data){
    beforeReadTimeDigitalIn(I2C_SLAVE_ADDR_LIST[1], 2, data);}

static void beforeReadTimeI2C1In3(UA_Server *server,
                                  const UA_NodeId *sessionId, void *sessionContext,
                                  const UA_NodeId *nodeid, void *nodeContext,
                                  const UA_NumericRange *range, const UA_DataValue *data){
    beforeReadTimeDigitalIn(I2C_SLAVE_ADDR_LIST[1], 3, data);}

// I2C2
static void beforeReadTimeI2C2In0(UA_Server *server,
                                  const UA_NodeId *sessionId, void *sessionContext,
                                  const UA_NodeId *nodeid, void *nodeContext,
                                  const UA_NumericRange *range, const UA_DataValue *data){
    beforeReadTimeDigitalIn(I2C_SLAVE_ADDR_LIST[2], 0, data);}

static void beforeReadTimeI2C2In1(UA_Server *server,
                                  const UA_NodeId *sessionId, void *sessionContext,
                                  const UA_NodeId *nodeid, void *nodeContext,
                                  const UA_NumericRange *range, const UA_DataValue *data){
    beforeReadTimeDigitalIn(I2C_SLAVE_ADDR_LIST[2], 1, data);}

static void beforeReadTimeI2C2In2(UA_Server *server,
                                  const UA_NodeId *sessionId, void *sessionContext,
                                  const UA_NodeId *nodeid, void *nodeContext,
                                  const UA_NumericRange *range, const UA_DataValue *data){
    beforeReadTimeDigitalIn(I2C_SLAVE_ADDR_LIST[2], 2, data);}

static void beforeReadTimeI2C2In3(UA_Server *server,
                                  const UA_NodeId *sessionId, void *sessionContext,
                                  const UA_NodeId *nodeid, void *nodeContext,
                                  const UA_NumericRange *range, const UA_DataValue *data){
    beforeReadTimeDigitalIn(I2C_SLAVE_ADDR_LIST[2], 3, data);}


// I2C3
static void beforeReadTimeI2C3In0(UA_Server *server,
                                  const UA_NodeId *sessionId, void *sessionContext,
                                  const UA_NodeId *nodeid, void *nodeContext,
                                  const UA_NumericRange *range, const UA_DataValue *data){
    beforeReadTimeDigitalIn(I2C_SLAVE_ADDR_LIST[3], 0, data);}

static void beforeReadTimeI2C3In1(UA_Server *server,
                                  const UA_NodeId *sessionId, void *sessionContext,
                                  const UA_NodeId *nodeid, void *nodeContext,
                                  const UA_NumericRange *range, const UA_DataValue *data){
    beforeReadTimeDigitalIn(I2C_SLAVE_ADDR_LIST[3], 1, data);}

static void beforeReadTimeI2C3In2(UA_Server *server,
                                  const UA_NodeId *sessionId, void *sessionContext,
                                  const UA_NodeId *nodeid, void *nodeContext,
                                  const UA_NumericRange *range, const UA_DataValue *data){
    beforeReadTimeDigitalIn(I2C_SLAVE_ADDR_LIST[3], 2, data);}

static void beforeReadTimeI2C3In3(UA_Server *server,
                                  const UA_NodeId *sessionId, void *sessionContext,
                                  const UA_NodeId *nodeid, void *nodeContext,
                                  const UA_NumericRange *range, const UA_DataValue *data){
    beforeReadTimeDigitalIn(I2C_SLAVE_ADDR_LIST[3], 3, data);}


// Relays
static void afterWriteTime(UA_Server *server,
                           const UA_NodeId *sessionId, void *sessionContext,
                           const UA_NodeId *nodeId, void *nodeContext,
                           const UA_NumericRange *range, const UA_DataValue *data)
{
    // nothing to do as not yet needed.
}

static void setI2CSlaveRelayState(int addr,
                                  int relay_number,
                                  uint8_t *global_relays_state,
                                  const UA_DataValue *data){
    /*
     * A generic handler function for setting relay state.
     */
    if (data->value.type == &UA_TYPES[UA_TYPES_INT32]){
        UA_Int32 hrValue = *(UA_Int32 *)data->value.data;
        // used only for debuging with logical analyzer
        if (CURRENT_GPIO_MODE == 2)
            setGPIO();

        if (hrValue > 0){
            *global_relays_state |= 1UL << relay_number;
            setRelayState(*global_relays_state, addr);
        }
        else{
            *global_relays_state &= ~(1UL << relay_number);
            setRelayState(*global_relays_state, addr);
        }
    }
}

// I2C0
static void afterWriteTimeI2C0_0(UA_Server *server,
                                 const UA_NodeId *sessionId, void *sessionContext,
                                 const UA_NodeId *nodeId, void *nodeContext,
                                 const UA_NumericRange *range, const UA_DataValue *data){
     setI2CSlaveRelayState(I2C_SLAVE_ADDR_LIST[0], 0, &I2C_0_RELAYS_STATE, data);}

static void afterWriteTimeI2C0_1(UA_Server *server,
                                 const UA_NodeId *sessionId, void *sessionContext,
                                 const UA_NodeId *nodeId, void *nodeContext,
                                 const UA_NumericRange *range, const UA_DataValue *data){
     setI2CSlaveRelayState(I2C_SLAVE_ADDR_LIST[0], 1, &I2C_0_RELAYS_STATE, data);}

static void afterWriteTimeI2C0_2(UA_Server *server,
                                 const UA_NodeId *sessionId, void *sessionContext,
                                 const UA_NodeId *nodeId, void *nodeContext,
                                 const UA_NumericRange *range, const UA_DataValue *data){
     setI2CSlaveRelayState(I2C_SLAVE_ADDR_LIST[0], 2, &I2C_0_RELAYS_STATE, data);}

static void afterWriteTimeI2C0_3(UA_Server *server,
                                 const UA_NodeId *sessionId, void *sessionContext,
                                 const UA_NodeId *nodeId, void *nodeContext,
                                 const UA_NumericRange *range, const UA_DataValue *data){
     setI2CSlaveRelayState(I2C_SLAVE_ADDR_LIST[0], 3, &I2C_0_RELAYS_STATE, data);}

// I2C1
static void afterWriteTimeI2C1_0(UA_Server *server,
                                 const UA_NodeId *sessionId, void *sessionContext,
                                 const UA_NodeId *nodeId, void *nodeContext,
                                 const UA_NumericRange *range, const UA_DataValue *data){
     setI2CSlaveRelayState(I2C_SLAVE_ADDR_LIST[1], 0, &I2C_1_RELAYS_STATE, data);}

static void afterWriteTimeI2C1_1(UA_Server *server,
                                 const UA_NodeId *sessionId, void *sessionContext,
                                 const UA_NodeId *nodeId, void *nodeContext,
                                 const UA_NumericRange *range, const UA_DataValue *data){
     setI2CSlaveRelayState(I2C_SLAVE_ADDR_LIST[1], 1, &I2C_1_RELAYS_STATE, data);}

static void afterWriteTimeI2C1_2(UA_Server *server,
                                 const UA_NodeId *sessionId, void *sessionContext,
                                 const UA_NodeId *nodeId, void *nodeContext,
                                 const UA_NumericRange *range, const UA_DataValue *data){
     setI2CSlaveRelayState(I2C_SLAVE_ADDR_LIST[1], 2, &I2C_1_RELAYS_STATE, data);}

static void afterWriteTimeI2C1_3(UA_Server *server,
                                 const UA_NodeId *sessionId, void *sessionContext,
                                 const UA_NodeId *nodeId, void *nodeContext,
                                 const UA_NumericRange *range, const UA_DataValue *data){
     setI2CSlaveRelayState(I2C_SLAVE_ADDR_LIST[1], 3, &I2C_1_RELAYS_STATE, data);}

// I2C2
static void afterWriteTimeI2C2_0(UA_Server *server,
                                 const UA_NodeId *sessionId, void *sessionContext,
                                 const UA_NodeId *nodeId, void *nodeContext,
                                 const UA_NumericRange *range, const UA_DataValue *data){
     setI2CSlaveRelayState(I2C_SLAVE_ADDR_LIST[2], 0, &I2C_2_RELAYS_STATE, data);}

static void afterWriteTimeI2C2_1(UA_Server *server,
                                 const UA_NodeId *sessionId, void *sessionContext,
                                 const UA_NodeId *nodeId, void *nodeContext,
                                 const UA_NumericRange *range, const UA_DataValue *data){
     setI2CSlaveRelayState(I2C_SLAVE_ADDR_LIST[2], 1, &I2C_2_RELAYS_STATE, data);}

static void afterWriteTimeI2C2_2(UA_Server *server,
                                 const UA_NodeId *sessionId, void *sessionContext,
                                 const UA_NodeId *nodeId, void *nodeContext,
                                 const UA_NumericRange *range, const UA_DataValue *data){
     setI2CSlaveRelayState(I2C_SLAVE_ADDR_LIST[2], 2, &I2C_2_RELAYS_STATE, data);}

static void afterWriteTimeI2C2_3(UA_Server *server,
                                 const UA_NodeId *sessionId, void *sessionContext,
                                 const UA_NodeId *nodeId, void *nodeContext,
                                 const UA_NumericRange *range, const UA_DataValue *data){
     setI2CSlaveRelayState(I2C_SLAVE_ADDR_LIST[2], 3, &I2C_2_RELAYS_STATE, data);}

// I2C3
static void afterWriteTimeI2C3_0(UA_Server *server,
                                 const UA_NodeId *sessionId, void *sessionContext,
                                 const UA_NodeId *nodeId, void *nodeContext,
                                 const UA_NumericRange *range, const UA_DataValue *data){
     setI2CSlaveRelayState(I2C_SLAVE_ADDR_LIST[3], 0, &I2C_3_RELAYS_STATE, data);}

static void afterWriteTimeI2C3_1(UA_Server *server,
                                 const UA_NodeId *sessionId, void *sessionContext,
                                 const UA_NodeId *nodeId, void *nodeContext,
                                 const UA_NumericRange *range, const UA_DataValue *data){
     setI2CSlaveRelayState(I2C_SLAVE_ADDR_LIST[3], 1, &I2C_3_RELAYS_STATE, data);}

static void afterWriteTimeI2C3_2(UA_Server *server,
                                 const UA_NodeId *sessionId, void *sessionContext,
                                 const UA_NodeId *nodeId, void *nodeContext,
                                 const UA_NumericRange *range, const UA_DataValue *data){
     setI2CSlaveRelayState(I2C_SLAVE_ADDR_LIST[3], 2, &I2C_3_RELAYS_STATE, data);}

static void afterWriteTimeI2C3_3(UA_Server *server,
                                 const UA_NodeId *sessionId, void *sessionContext,
                                 const UA_NodeId *nodeId, void *nodeContext,
                                 const UA_NumericRange *range, const UA_DataValue *data){
     setI2CSlaveRelayState(I2C_SLAVE_ADDR_LIST[3], 3, &I2C_3_RELAYS_STATE, data);}


static void addValueCallbackToCurrentTimeVariable(UA_Server *server)
{
    int device_slave_list_length = getI2CSlaveListLength();
    int device_number, offset_number;
    char *node_id = malloc(sizeof(char) * 10);

    // list of relays, digital and analog callbacks for all devices (XXX: for now only 4 outof 127 possible!)
    static void (*callables[48])
                                 (UA_Server *server, const UA_NodeId *sessionId, void *sessionContext,
                                  const UA_NodeId *nodeId, void *nodeContext, const UA_NumericRange *range,
                                  const UA_DataValue *data) = {// relays - IC0
                                                               afterWriteTimeI2C0_0,
                                                               afterWriteTimeI2C0_1,
                                                               afterWriteTimeI2C0_2,
                                                               afterWriteTimeI2C0_3,
                                                               // digital - IC0
                                                               beforeReadTimeI2C0In0,
                                                               beforeReadTimeI2C0In1,
                                                               beforeReadTimeI2C0In2,
                                                               beforeReadTimeI2C0In3,
                                                               // analong - IC0
                                                               beforeReadTimeI2C0Ain0,
                                                               beforeReadTimeI2C0Ain1,
                                                               beforeReadTimeI2C0Ain2,
                                                               beforeReadTimeI2C0Ain3,

                                                               // relays - IC1
                                                               afterWriteTimeI2C1_0,
                                                               afterWriteTimeI2C1_1,
                                                               afterWriteTimeI2C1_2,
                                                               afterWriteTimeI2C1_3,
                                                               // digital - IC1
                                                               beforeReadTimeI2C1In0,
                                                               beforeReadTimeI2C1In1,
                                                               beforeReadTimeI2C1In2,
                                                               beforeReadTimeI2C1In3,
                                                               // analog - IC1
                                                               beforeReadTimeI2C1Ain0,
                                                               beforeReadTimeI2C1Ain1,
                                                               beforeReadTimeI2C1Ain2,
                                                               beforeReadTimeI2C1Ain3,

                                                               // relays - IC2
                                                               afterWriteTimeI2C2_0,
                                                               afterWriteTimeI2C2_1,
                                                               afterWriteTimeI2C2_2,
                                                               afterWriteTimeI2C2_3,
                                                               // digital - IC2
                                                               beforeReadTimeI2C2In0,
                                                               beforeReadTimeI2C2In1,
                                                               beforeReadTimeI2C2In2,
                                                               beforeReadTimeI2C2In3,
                                                               // analong - IC2
                                                               beforeReadTimeI2C2Ain0,
                                                               beforeReadTimeI2C2Ain1,
                                                               beforeReadTimeI2C2Ain2,
                                                               beforeReadTimeI2C2Ain3,

                                                               // relays - IC3
                                                               afterWriteTimeI2C3_0,
                                                               afterWriteTimeI2C3_1,
                                                               afterWriteTimeI2C3_2,
                                                               afterWriteTimeI2C3_3,
                                                               // digital - IC3
                                                               beforeReadTimeI2C3In0,
                                                               beforeReadTimeI2C3In1,
                                                               beforeReadTimeI2C3In2,
                                                               beforeReadTimeI2C3In3,
                                                               // analog - IC1
                                                               beforeReadTimeI2C3Ain0,
                                                               beforeReadTimeI2C3Ain1,
                                                               beforeReadTimeI2C3Ain2,
                                                               beforeReadTimeI2C3Ain3,
                                };

    // set proper callbacks from list of all
    for (device_number = 0; device_number < device_slave_list_length;  device_number++){
        for (offset_number = 0; offset_number < 4; offset_number++){
            // relays
            sprintf(node_id, "i2c%d.relay%d", device_number, offset_number);
            UA_NodeId currentNodeId = UA_NODEID_STRING(1, node_id);
            UA_ValueCallback callback;
            callback.onRead = beforeReadTime;
            callback.onWrite = callables[device_number * 12 + offset_number];
            UA_Server_setVariableNode_valueCallback(server, currentNodeId, callback);

            // digital inputs
            sprintf(node_id, "i2c%d.in%d", device_number, offset_number);
            UA_NodeId currentNodeId1 = UA_NODEID_STRING(1, node_id);
            UA_ValueCallback callback1;
            callback1.onRead = callables[device_number * 12 + offset_number + 4];
            callback1.onWrite = afterWriteTime;
            UA_Server_setVariableNode_valueCallback(server, currentNodeId1, callback1);

            // analog inputs
            sprintf(node_id, "i2c%d.ain%d", device_number, offset_number);
            UA_NodeId currentNodeId2 = UA_NODEID_STRING(1, node_id);
            UA_ValueCallback callback2;
            callback2.onRead = callables[device_number * 12 + offset_number + 8];
            callback2.onWrite = afterWriteTime;
            UA_Server_setVariableNode_valueCallback(server, currentNodeId2, callback2);
    }
 }
}
